@@ -35,21 +35,9 @@ typedef unsigned long long u64;
 #error I need at least some 64-bit type
 #endif
 
-#ifdef NBD_H_LOCAL
-/* 2.6.18 and above use __be* rather than u* */
 #define __be32 u32
 #define __be64 u64
 #include "nbd.h"
-#endif
-#ifdef NBD_H_LINUX
-# ifdef HAVE_LINUX_TYPES_H
-#  include <linux/types.h>
-# else
-#  define __be32 u32
-#  define __be64 u64
-# endif
-# include <linux/nbd.h>
-#endif
 
 #if NBD_LFS==1
 #define _LARGEFILE_SOURCE
@@ -57,6 +45,7 @@ typedef unsigned long long u64;
 #endif
 
 u64 cliserv_magic = 0x00420281861253LL;
+u64 opts_magic = 0x49484156454F5054LL;
 #define INIT_PASSWD "NBDMAGIC"
 
 #define INFO(a) do { } while(0)
@@ -87,13 +76,10 @@ void setmysockopt(int sock) {
 #endif
 #endif
 
-void err(const char *s) G_GNUC_NORETURN;
+void err_nonfatal(const char *s) {
+	char s1[150], *s2;
 
-void err(const char *s) {
-	const int maxlen = 150;
-	char s1[maxlen], *s2;
-
-	strncpy(s1, s, maxlen);
+	strncpy(s1, s, sizeof(s1));
 	if ((s2 = strstr(s, "%m"))) {
 		strcpy(s1 + (s2 - s), strerror(errno));
 		s2 += 2;
@@ -108,13 +94,19 @@ void err(const char *s) {
 	}
 #endif
 
-	s1[maxlen-1] = '\0';
+	s1[sizeof(s1)-1] = '\0';
 #ifdef ISSERVER
 	syslog(LOG_ERR, "%s", s1);
 	syslog(LOG_ERR, "Exiting.");
 #endif
 	fprintf(stderr, "Error: %s\nExiting.\n", s1);
-	exit(1);
+}
+
+void err(const char *s) G_GNUC_NORETURN;
+
+void err(const char *s) {
+	err_nonfatal(s);
+	exit(EXIT_FAILURE);
 }
 
 void logging(void) {
@@ -143,3 +135,9 @@ u64 ntohll(u64 a) {
 /* Flags used between the client and server */
 #define NBD_FLAG_HAS_FLAGS	(1 << 0)	/* Flags are there */
 #define NBD_FLAG_READ_ONLY	(1 << 1)	/* Device is read-only */
+
+#define NBD_DEFAULT_PORT	"10809"	/* Port on which named exports are
+					 * served */
+
+/* Options that the client can select to the server */
+#define NBD_OPT_EXPORT_NAME	(1 << 0)	/* Client wants to select a named export (is followed by length and name of export) */
